var util = require('util');

var through = require('through'),
    logref = require('logref'),
    funstance = require('funstance'),
    colorize = require('./colorize');

var loggins = module.exports = createLogger();

loggins.colorize = colorize;

function createLogger(opts) {
  opts = opts || {};

  //
  // "pretty" messages are readable from logger
  //
  var logger = funstance(through(), createLogger);

  //
  // Levels are used to auto-attach log methods.
  // Colors are used by default _pretty implementation.
  //
  var lvls, colors;

  lvls = opts.levels || [
    'debug',
    'info',
    'warn',
    'error'
  ];

  //
  // Levels are a hash? Values must be colors.
  //
  if (!Array.isArray(lvls) && typeof lvls === 'object') {
    colors = lvls;
    lvls = Object.keys(lvls);
  }

  if (!colors) {
    colors = {
      debug: 'grey',
      info: 'green',
      warn: 'yellow',
      error: 'red'
    };
  }

  //
  // JSON-stringified messages are readable from logger.raw
  //
  logger.raw = through();

  //
  // Punch this to customize what pretty messages look like
  //
  logger._pretty = function (o) {
    var lvl = o.level,
        msg = o.message;

    if (logger.colors[lvl]) {
      //
      // Associated color? Colorize the level.
      //
      lvl = loggins.colorize(lvl, logger.colors[lvl]);
    }

    return msg
      .replace(/\n$/, '')
      .split('\n')
      .map(function (l) {
        return lvl + ': ' + l;
      }).join('\n')
      + '\n'
    ;
  };

  logger.colors = colors;

  //
  // Call as a logging method, also returns a stream instance
  //
  // Write: logging messages at that lvl (raw)
  // Read: messages at that log level
  // call: Write with a util.format.apply
  //
  logger.log = function (lvl) {
    if (logger[lvl]) {
      return logger[lvl];
    }

    var logStream = through(function write(buff) {
      var message = Buffer.isBuffer(buff) ? buff.toString() : buff;
      this.queue(message);
    });

    function write(message) {
      var o = { level: lvl, message: message, timestamp: Date.now() };

      logger.write(logger._pretty(o));
      logStream.emit('data', message);
      return logger.raw.write(JSON.stringify(o));
    }

    return logger[lvl] = funstance(logStream, function logger() {
      return write(util.format.apply(null, arguments) + '\n');
    });
  };

  //
  // Level-based loggers
  //
  lvls.forEach(function (lvl) {
    logger.log(lvl);
  });

  //
  // logref support
  //
  logger.logging = function logging(name) {
    var lr = logref(name),
        l = logger.log(name);

    lr.logger.on('msg', function (msg) {
      l(msg);
    });

    return funstance(l, function write(msg, ctx) {
      lr(msg, ctx || {});
    });
  }

  return logger;
};
